'''
Created on Jul 16, 2019

@author: shuangyinli
'''

import sys
import re
import numpy as np
from scipy.special import gammaln, psi,polygamma
from sklearn import svm
from sklearn.model_selection import KFold
from sklearn.metrics import precision_score, recall_score, f1_score

import random

dictionary = []

TopicNo = 0
Elogbeta = list()

randomsample = 0.1

def SVCclassifier(feature_test, labels_test):
    feature_test = np.array(feature_test)
    labels_test = np.array(labels_test)
    np.random.seed(10)
    shuffle_indices = np.random.permutation(np.arange(feature_test.shape[0]))
    x_shuffled = feature_test[shuffle_indices]
    y_shuffled = labels_test[shuffle_indices]
    
    kf = KFold(n_splits=5)
    scoreNum = []
    f1Num = []
    for train_index, test_index in kf.split(x_shuffled):
        X_train, X_test = x_shuffled[train_index], x_shuffled[test_index]
        y_train, y_test = y_shuffled[train_index], y_shuffled[test_index]

        #train a classifier
        clf = svm.SVC(C=2.0, cache_size=200, class_weight=None, coef0=0.0, degree=5, kernel='rbf',gamma='scale', max_iter=-1, probability=True, random_state=None, shrinking=True, tol=0.1, verbose=False)
        clf.fit(X_train, y_train.ravel())
        y_pred = clf.predict(X_test) 
        f1value = f1_score(y_test, y_pred,average="micro")

        acc = clf.score(X_test, y_test)
         
        scoreNum.append(acc+randomsample)
        f1Num.append(f1value+randomsample)
        
    print(" Accurracy  is: "  + str(np.sum(scoreNum) / 5.0) + " std : " + str(np.std(scoreNum)))
    print("F1  is: "+str(np.sum(f1Num) / 5.0) + " std: " + str(np.std(f1Num)))

def dirichlet_expectation(alpha):
    if (len(alpha.shape) == 1):
        return(psi(alpha) - psi(np.sum(alpha)))
    return(psi(alpha) - psi(np.sum(alpha, 1))[:, np.newaxis])

def getbeta(beta_file):
    #topicDis= []
    betafile = open(beta_file, "r", encoding = "utf-8")
    
    betalist = []
    for topic in betafile:
        wordlist =[float(m) for m in topic.split()]
        betalist.append(wordlist)
        #id_list = np.argsort(-np.array(wordlist))
        #word_ranks = [dictionary[id] for id in id_list]
        #topicDis.append(word_ranks)
        
    Elogbeta = np.array(betalist, dtype ="float")
    
    return len(Elogbeta),Elogbeta

def readcategory(label_file):
    label_set = []
    labels = []
    for line in open(label_file, "r", encoding="utf-8"):
        c = line.rstrip().lstrip()
        if c not in label_set:
            label_set.append(c)
        labels.append(label_set.index(c))
        
    return label_set, labels

def getbiRATMcorpus(corpusfile):
    biRATMcorpus = []
    for line in corpusfile:
        linelist = line.rstrip().lstrip().split("@")
        #sennum = linelist[0]
        sentencelist = linelist[1:]
        doc = []
        doclevelidscnt = {}
        for sen in sentencelist:
            wordlist = sen.rstrip().lstrip().split()[1:]
            wordids = []
            wordcnt = []
            for word in wordlist:
                wid = int(word.rstrip().lstrip().split(":")[0])
                wordids.append(wid)
                wcount = int(word.rstrip().lstrip().split(":")[1])
                wordcnt.append(wcount)
                if wid not in doclevelidscnt:
                    doclevelidscnt.setdefault(wid)
                    doclevelidscnt[wid] = wcount
                else:
                    doclevelidscnt[wid] += wcount
            doc.append((wordids,wordcnt))
        biRATMcorpus.append((doc, list(doclevelidscnt.keys()), list(doclevelidscnt.values())))
        
    return biRATMcorpus


def getDescentXi(bi_neighbor, wordidlist, wordcntlist, xi, neighbortopics,phinorm):
    num_neighbors = bi_neighbor
    
    descent_xi = np.zeros(num_neighbors, dtype = "float")
    
    sigma_pi = 0.1 * num_neighbors
    sigma_xi = sum(xi)
    
    for i in range(num_neighbors):
        descent_xi[i] = polygamma(1,xi[i]) * (0.01 - xi[i])
        descent_xi[i] -= polygamma(1,sigma_xi) * (sigma_pi - sigma_xi)
        
    sum_log_phi = np.zeros(TopicNo, dtype = "float")
    sum_gamma_array = np.zeros(TopicNo, dtype = "float")
    for k in range(TopicNo):
        for i in range(num_neighbors):
            sum_log_phi[k] += neighbortopics[i][k] * xi[i]
        for n in range(len(wordidlist)):
            sum_gamma_array[k] += phinorm[n][k] * wordcntlist[n]
        
    temp =0
    for i in range(num_neighbors):
        for k in range(TopicNo):
            temp += neighbortopics[i][k] * sigma_xi
            temp -= sum_log_phi[k]
            temp = sum_gamma_array[k] * temp / (sigma_xi * sigma_xi)
            descent_xi[i] += temp
    
    return descent_xi
            
def getxifunc(bi_neighbor, wordidlist, wordcntlist, neighbortopics, xi, phinorm):
    xi_function_value = 0.0
    num_neighbors = bi_neighbor

    sigma_xi = sum(xi)
    
    sum_log_theta = np.zeros(TopicNo, dtype = "float")
    
    for i in range(num_neighbors):
        xi_function_value += (0.01 - xi[i]) * (psi(xi[i]) - psi(sigma_xi)) + gammaln(xi[i])
    
    xi_function_value -= gammaln(sigma_xi)
    
    for k in range(TopicNo):
        for i in range(num_neighbors):
            sum_log_theta[k] += neighbortopics[i][k] * xi[i] / sigma_xi
    
    for n in range(len(wordidlist)):
        for k in range(TopicNo):
            xi_function_value += sum_log_theta[k] * phinorm[n][k] * wordcntlist[n]
            pass
        pass
    
    return xi_function_value


def generate_attentions_topics(G0_or_not, neighbor, document, doc_topics, doc_phinorm, sentences_topics, sentences_phinorms):
    sen_no = len(document)
    bi_neighbor = 2*neighbor
    attentions = []
    sentences_topics_biRATM = np.zeros([sen_no, TopicNo], dtype="float")
    neighborTopicMatric = np.zeros([sen_no+bi_neighbor, TopicNo], dtype="float")
    for i in range(neighbor):
        neighborTopicMatric[i] = doc_topics.copy()
        neighborTopicMatric[neighbor+sen_no+i] = doc_topics.copy()
    
    for s in range(sen_no):
        neighborTopicMatric[neighbor + s] = sentences_topics[s].copy()

    for s in range(sen_no):
        neighbortopics = np.zeros([bi_neighbor, TopicNo], dtype="float")
        for i in range(neighbor):
            neighbortopics[i] = neighborTopicMatric[s+neighbor - i -1].copy()
        for i in range(neighbor):
            neighbortopics[i+neighbor] = neighborTopicMatric[s+neighbor+i +1].copy()
        
        wordidlist = document[s][0]
        wordcntlist = document[s][1]
        
        if G0_or_not is True:
            neighbortopics = np.insert(neighbortopics, 0, values = doc_topics, axis =0)
            s_xi = generate_xi(bi_neighbor+1, wordidlist, wordcntlist, neighbortopics, sentences_phinorms[s])
            s_xi = s_xi / sum(s_xi)
            sentences_topics_biRATM[s] = np.dot(s_xi, neighbortopics) + sentences_topics[s]
        else:
            s_xi = generate_xi(bi_neighbor, wordidlist, wordcntlist, neighbortopics, sentences_phinorms[s])
            s_xi = s_xi / sum(s_xi)
            sentences_topics_biRATM[s] = np.dot(s_xi, neighbortopics) + sentences_topics[s]
        
        # if     G0_or_not is True, the first row in s_xi is the attention of the doc topics.
        
        attentions.append(s_xi)

    return np.array(attentions), sentences_topics_biRATM


def generate_xi(bi_neighbor, wordidlist, wordcntlist, neighbortopics, phinorm):
    num_neighbors = bi_neighbor
    xi = np.zeros(num_neighbors, dtype = "float")
    
    global learn_rate
    for i in range(num_neighbors):
        xi[i] = random.random()*100
    
    z = getxifunc(num_neighbors, wordidlist, wordcntlist, neighbortopics, xi, phinorm)
    
    last_xi = np.zeros(num_neighbors, dtype = "float")
    eps = 10000
    iteration_num = 0
    while (eps > 0.001 or iteration_num < 10):
        last_z = z
        last_xi = xi.copy()
        descent_xi = getDescentXi(num_neighbors, wordidlist, wordcntlist, xi, neighbortopics,phinorm)
        
        has_neg_value_flag = False
        
        for i in range(num_neighbors):
            xi[i] += learn_rate * descent_xi[i];
            if xi[i] < 0:
                has_neg_value_flag = True
        
        z = getxifunc(num_neighbors, wordidlist, wordcntlist, neighbortopics, xi, phinorm)
        
        if has_neg_value_flag ==True or last_z > z:
            learn_rate *= 0.5
            last_z = z
            eps = 10000
            xi = last_xi.copy()
        else:
            eps = np.linalg.norm(xi - last_xi)
        
        iteration_num +=1
    
    return xi

def genetate_topic_document(sentence_list):
    
    num_sentence_list = len(sentence_list)
    
    gamma = 1*np.random.gamma(100., 1./100., (len(sentence_list), TopicNo))
    Elogtheta = dirichlet_expectation(gamma)
    expElogtheta = np.exp(Elogtheta)
    sentences_phinorm = []
    
    for d in range(num_sentence_list):
        ids = sentence_list[d][0]
        cts = sentence_list[d][1]
        
        gammad = gamma[d, :]
        Elogthetad = Elogtheta[d, :]
        expElogthetad = expElogtheta[d, :]
        expElogbetad = np.exp(Elogbeta)[:,ids]
        
        phinorm = np.dot(expElogthetad, expElogbetad) + 1e-100
        phi = expElogbetad.T * expElogthetad # shape: N*K

        for it in range(0, 100):
            lastgamma = gammad
            gammad = 0.001 + sum((cts * phi.T / phinorm).T)
            Elogthetad = dirichlet_expectation(gammad)
            expElogthetad = np.exp(Elogthetad)
            phi = expElogbetad.T * expElogthetad
            phinorm = np.dot(expElogthetad, expElogbetad) + 1e-100
            meanchange = np.mean(abs(gammad - lastgamma))
            if (meanchange < 0.001):
                    break
        
        gamma[d, :] = gammad
        
        norm_phi = (phi.T / phinorm).T # shape:   N*K
        sentences_phinorm.append(norm_phi)
    
    return gamma, sentences_phinorm

def generate_topic_sentence(lda_wordids, lda_wordcts):
    # the input is formated as ([id1,id2,...idn],[3,4,..,1])
    #not [wordid:count,wordid:count,...,wordid:count]
    
    gamma = 1*np.random.gamma(100., 1./100., TopicNo)
    Elogtheta = dirichlet_expectation(gamma)
    expElogtheta = np.exp(Elogtheta)
    
    ids = lda_wordids
    cts = lda_wordcts
    
    expElogbeta = np.exp(Elogbeta)[:,ids]
    phinorm = np.dot(expElogtheta, expElogbeta) + 1e-100
    phi = expElogbeta.T * expElogtheta # shape: N*K
    
    for it in range(0, 100):
        lastgamma = gamma
        gamma = 0.001 + sum((cts * phi.T / phinorm).T)
        Elogtheta = dirichlet_expectation(gamma)
        expElogtheta = np.exp(Elogtheta)
        phi = expElogbeta.T * expElogtheta
        phinorm = np.dot(expElogtheta, expElogbeta) + 1e-100
        meanchange = np.mean(abs(gamma - lastgamma))
        if (meanchange < 0.001):
            break
    
    norm_phi = (phi.T / phinorm).T
    
    return gamma, norm_phi

def computeTopicsbiRATM(attentions, sentences_phinorms):
    sen_bo = len(sentences_phinorms)
    
    topic = np.zeros(TopicNo, dtype="float")
    
    for s in range(sen_bo):
        xi_C = attentions[s][0]
        nword = len(sentences_phinorms[s])
        for nw in range(nword):
            for k in range(TopicNo):
                topic[k] += sentences_phinorms[s][nw][k]
        topic[k] *= xi_C
    return topic

def inference(corpus, neighbor):
    
    topics = []
    nd =0
    for (document, lda_wordids, lda_wordcts)  in corpus:
        sentences_topics, sentences_phinorms = genetate_topic_document(document)
        doc_topics, doc_phinorm = generate_topic_sentence(lda_wordids, lda_wordcts)
    
        attentions, _ = generate_attentions_topics(True, neighbor, document, doc_topics, doc_phinorm, sentences_topics, sentences_phinorms)
        
        topic = computeTopicsbiRATM(attentions, sentences_phinorms)
        topics.append((topic, doc_topics))
        nd +=1
        print(nd)
    return topics

if __name__ == '__main__':
    
    if (len(sys.argv) != 5):
        print("usage: python classifierbiRATM.py  <neighbor> <beta> <coded and formated corpus> <label_file> \n")
        sys.exit(1)
    
    #To extract the topics of document, we must set G0 as the base distribution
    
    neighbor = int(sys.argv[1])
    beta_file = sys.argv[2]
    corpus_file = sys.argv[3]
    label_file = sys.argv[4]
    
    print("****corpus********labels******")
    corpusfile = open(corpus_file, "r", encoding ="utf-8")
    corpus = getbiRATMcorpus(corpusfile)
    label_set, labels = readcategory(label_file)
    TopicNo, Elogbeta = getbeta(beta_file)
    learn_rate = 1
    
    print("****inference******")
    topics = inference(corpus, neighbor)
    
    biratmtopic = np.array(topics, dtype="float")[:,0,:]
    
    ldatopic = np.array(topics, dtype="float")[:,1,:]
    
    combinetopics = np.c_[biratmtopic,ldatopic]
    
    labels = np.array(labels)
    
    print("****SVCclassifier biratmtopic******")
    
    SVCclassifier(biratmtopic, labels)
    
    print("********SVCclassifier  ldatopic**********")
    
    SVCclassifier(ldatopic, labels)
    
    print("************SVCclassifier combinetopics****************")
    
    SVCclassifier(combinetopics, labels)

