'''
Created on Jul 5, 2019

@author: shuangyinli
'''

#!/usr/bin/python
# -*- coding: utf-8 -*-
 
import torch
from torch import nn
import numpy as np
from torch.autograd import Variable
import torch.nn.functional as F
 
class TextRNN(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, output_size_, dictionary_size):
        super(TextRNN, self).__init__()
    
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)  # 进行词嵌入
        self.rnn = nn.RNN(input_size=input_size_, hidden_size=hidden_size_, num_layers=1)
        self.f1 = nn.Sequential(nn.Linear(hidden_size_,hidden_size_),
                                nn.Dropout(0.5),
                                nn.ReLU())
        self.f2 = nn.Sequential(nn.Linear(hidden_size_,output_size_), nn.Softmax(dim=1))
 
    def forward(self, x):
        x = self.embedding(x)
        #x = F.embedding(x, self.embedding_matrix)
        x,_ = self.rnn(x)
        x = self.f1(x[:,-1,:])
        return self.f2(x)

class TextRNNSeqToSeq(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, dictionary_size):
        super(TextRNNSeqToSeq, self).__init__()
        
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)  # 进行词嵌入
        self.encoder = nn.Sequential(nn.RNN(input_size=input_size_, hidden_size=hidden_size_, num_layers=1))
        self.decoder = nn.Sequential(nn.RNN(input_size=hidden_size_, hidden_size=input_size_, num_layers=1))
    
    def forward(self, x):
        x_embedding = self.embedding(x)
        encodding,_ = self.encoder(x_embedding)
        decodding,_ = self.decoder(encodding)
        return x_embedding, encodding, decodding

class TextLSTMSeqToSeq(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, dictionary_size):
        super(TextLSTMSeqToSeq, self).__init__()
        
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)  # 进行词嵌入
        self.encoder = nn.Sequential(nn.LSTM(input_size=input_size_, hidden_size=hidden_size_, num_layers=1))
        self.decoder = nn.Sequential(nn.LSTM(input_size=hidden_size_, hidden_size=input_size_, num_layers=1))
    
    def forward(self, x):
        x_embedding = self.embedding(x)
        encodding,_ = self.encoder(x_embedding)
        decodding,_ = self.decoder(encodding)
        return x_embedding, encodding, decodding
    
class TextbiLSTMSeqToSeq(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, dictionary_size):
        super(TextbiLSTMSeqToSeq, self).__init__()
        
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=2*input_size_)  # 进行词嵌入
        self.encoder = nn.Sequential(nn.LSTM(input_size=2*input_size_, hidden_size=hidden_size_, num_layers=1,bidirectional=True))
        self.decoder = nn.Sequential(nn.LSTM(input_size=2*hidden_size_, hidden_size=input_size_, num_layers=1,bidirectional=True))
    
    def forward(self, x):
        x_embedding = self.embedding(x)
        encodding,_ = self.encoder(x_embedding)
        decodding,_ = self.decoder(encodding)
        return x_embedding, encodding, decodding
    
class TextGRUSeqToSeq(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, dictionary_size):
        super(TextGRUSeqToSeq, self).__init__()
        
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)  # 进行词嵌入
        self.encoder = nn.Sequential(nn.GRU(input_size=input_size_, hidden_size=hidden_size_, num_layers=1))
        self.decoder = nn.Sequential(nn.GRU(input_size=hidden_size_, hidden_size=input_size_, num_layers=1))
    
    def forward(self, x):
        x_embedding = self.embedding(x)
        encodding,_ = self.encoder(x_embedding)
        decodding,_ = self.decoder(encodding)
        return x_embedding, encodding, decodding

class TextLSTM(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, output_size_, dictionary_size):
        super(TextLSTM, self).__init__()

        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)
        
        self.rnn = nn.LSTM(input_size=input_size_, hidden_size=hidden_size_, num_layers=1, bidirectional=False)
        self.f1 = nn.Sequential(nn.Linear(hidden_size_,hidden_size_),
                                nn.Dropout(0.5),
                                nn.ReLU())
        self.f2 = nn.Sequential(nn.Linear(hidden_size_,output_size_), nn.Softmax(dim=1))
 
    def forward(self, x):
        x = self.embedding(x)
        #x = F.embedding(x, self.embedding_matrix)
        x,_ = self.rnn(x)
        #x = F.dropout(x,p=0.8)
        x = self.f1(x[:,-1,:])
        return self.f2(x)

class TextbiLSTM(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, output_size_, dictionary_size):
        super(TextbiLSTM, self).__init__()
        
        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)
        
        self.rnn = nn.LSTM(input_size=input_size_, hidden_size=hidden_size_, num_layers=1, bidirectional=True)
        self.f1 = nn.Sequential(nn.Linear(2*hidden_size_,hidden_size_),
                                nn.Dropout(0.5),
                                nn.ReLU())
        self.f2 = nn.Sequential(nn.Linear(hidden_size_,output_size_), nn.Softmax(dim=1))
 
    def forward(self, x):
        x = self.embedding(x)
        #x = F.embedding(x, self.embedding_matrix)
        x,_ = self.rnn(x)
        #x = F.dropout(x,p=0.8)
        x = self.f1(x[:,-1,:])
        return self.f2(x)

class TextGRU(nn.Module):
    """文本分类，RNN模型"""
    def __init__(self, input_size_, hidden_size_, output_size_, dictionary_size):
        super(TextGRU, self).__init__()

        self.embedding = nn.Embedding(num_embeddings=dictionary_size, embedding_dim=input_size_)
        
        self.rnn = nn.GRU(input_size=input_size_, hidden_size=hidden_size_, num_layers=1, bidirectional=True)
        self.f1 = nn.Sequential(nn.Linear(2*hidden_size_,hidden_size_),
                                nn.Dropout(0.5),
                                nn.ReLU())
        self.f2 = nn.Sequential(nn.Linear(hidden_size_,output_size_), nn.Softmax(dim=1))
 
    def forward(self, x):
        x = self.embedding(x)
        #x = F.embedding(x, self.embedding_matrix)
        x,_ = self.rnn(x)
        #x = F.dropout(x,p=0.8)
        x = self.f1(x[:,-1,:])
        return self.f2(x)

class TextCNN(nn.Module):
    def __init__(self):
        super(TextCNN, self).__init__()
        self.embedding = nn.Embedding(5000,64)
        self.conv = nn.Conv1d(64,256,5)
        self.f1 = nn.Sequential(nn.Linear(256*596, 128),
                                nn.ReLU())
        self.f2 = nn.Sequential(nn.Linear(128, 10),
                                nn.Softmax())
    def forward(self, x):
        x = self.embedding(x)
        x = x.detach().numpy()
        x = np.transpose(x,[0,2,1])
        x = torch.Tensor(x)
        x = Variable(x)
        x = self.conv(x)
        x = x.view(-1,256*596)
        x = self.f1(x)
        return self.f2(x)